{%- if show_source and has_source and sourcename %}
  <script>
    document.addEventListener('DOMContentLoaded', () => {
        const title = document.querySelector('meta[property="og:title"]').content;
        const elements = document.querySelectorAll('.improvepage');
        const pageurl = window.location.href.split('?')[0];
        elements.forEach(element => {
            const url = new URL(element.href.split('?')[0].replace("-nojs", ""));
            url.searchParams.set('pagetitle', title);
            url.searchParams.set('pageurl', pageurl);
            url.searchParams.set('pagesource', "{{ pagename }}.rst");
            element.href = url.toString();
        });
    });
  </script>
  <div role="note" aria-label="source link">
    <h3>{{ _('This page') }}</h3>
    <ul class="this-page-menu">
      <li><a href="{{ pathto('bugs') }}">{% trans %}Report a bug{% endtrans %}</a></li>
      <li><a class="improvepage" href="{{ pathto('improve-page-nojs') }}">{% trans %}Improve this page{% endtrans %}</a></li>
      <li>
        <a href="https://github.com/python/cpython/blob/main/Doc/{{ pagename }}.rst?plain=1"
            rel="nofollow">{{ _('Show source') }}
        </a>
      </li>
      {% if language != "en" %}
      <li>
        <a href="https://github.com/python/python-docs-{{ language }}/blob/{{ version }}/{{ pagename }}.po?plain=1"
           rel="nofollow">{{ _('Show translation source') }}</a>
      </li>
      {% endif %}
    </ul>
  </div>
{%- endif %}
