/*
 * Secret Labs' Regular Expression Engine
 *
 * regular expression matching engine
 *
 * Auto-generated by Tools/build/generate_sre_constants.py from
 * Lib/re/_constants.py.
 *
 * Copyright (c) 1997-2001 by Secret Labs AB.  All rights reserved.
 *
 * See the sre.c file for information on usage and redistribution.
 */

static void *sre_targets[43] = {
    &&TARGET_SRE_OP_FAILURE,
    &&TARGET_SRE_OP_SUCCESS,
    &&TARGET_SRE_OP_ANY,
    &&TARGET_SRE_OP_ANY_ALL,
    &&TARGET_SRE_OP_ASSERT,
    &&TARGET_SRE_OP_ASSERT_NOT,
    &&TARGET_SRE_OP_AT,
    &&TARGET_SRE_OP_BRANCH,
    &&TARGET_SRE_OP_CATEGORY,
    &&TARGET_SRE_OP_CHARSET,
    &&TARGET_SRE_OP_BIGCHARSET,
    &&TARGET_SRE_OP_GROUPREF,
    &&TARGET_SRE_OP_GROUPREF_EXISTS,
    &&TARGET_SRE_OP_IN,
    &&TARGET_SRE_OP_INFO,
    &&TARGET_SRE_OP_JUMP,
    &&TARGET_SRE_OP_LITERAL,
    &&TARGET_SRE_OP_MARK,
    &&TARGET_SRE_OP_MAX_UNTIL,
    &&TARGET_SRE_OP_MIN_UNTIL,
    &&TARGET_SRE_OP_NOT_LITERAL,
    &&TARGET_SRE_OP_NEGATE,
    &&TARGET_SRE_OP_RANGE,
    &&TARGET_SRE_OP_REPEAT,
    &&TARGET_SRE_OP_REPEAT_ONE,
    &&TARGET_SRE_OP_SUBPATTERN,
    &&TARGET_SRE_OP_MIN_REPEAT_ONE,
    &&TARGET_SRE_OP_ATOMIC_GROUP,
    &&TARGET_SRE_OP_POSSESSIVE_REPEAT,
    &&TARGET_SRE_OP_POSSESSIVE_REPEAT_ONE,
    &&TARGET_SRE_OP_GROUPREF_IGNORE,
    &&TARGET_SRE_OP_IN_IGNORE,
    &&TARGET_SRE_OP_LITERAL_IGNORE,
    &&TARGET_SRE_OP_NOT_LITERAL_IGNORE,
    &&TARGET_SRE_OP_GROUPREF_LOC_IGNORE,
    &&TARGET_SRE_OP_IN_LOC_IGNORE,
    &&TARGET_SRE_OP_LITERAL_LOC_IGNORE,
    &&TARGET_SRE_OP_NOT_LITERAL_LOC_IGNORE,
    &&TARGET_SRE_OP_GROUPREF_UNI_IGNORE,
    &&TARGET_SRE_OP_IN_UNI_IGNORE,
    &&TARGET_SRE_OP_LITERAL_UNI_IGNORE,
    &&TARGET_SRE_OP_NOT_LITERAL_UNI_IGNORE,
    &&TARGET_SRE_OP_RANGE_UNI_IGNORE,
};
